from openfisca_uk.api import *
from openfisca_uk.graphs import waterfall_chart


def test_single_reform_waterfall_chart_runs():
    reform = reforms.structural.abolish(
        "personal_allowance"
    ), reforms.parametric.set_parameter(
        BASELINE_PARAMETERS.benefit.child_benefit.amount.eldest, 50
    )
    waterfall_chart(reform, ["Abolish PA", "Increase CB"])


def test_multiple_reform_waterfall_chart_runs():
    reform_1 = reforms.structural.abolish(
        "personal_allowance"
    ), reforms.parametric.set_parameter(
        BASELINE_PARAMETERS.benefit.child_benefit.amount.eldest, 50
    )
    reform_2 = reforms.structural.abolish(
        "personal_allowance"
    ), reforms.parametric.set_parameter(
        BASELINE_PARAMETERS.benefit.child_benefit.amount.eldest, 300
    )
    waterfall_chart(
        [reform_1, reform_2],
        ["Abolish PA", "Increase CB"],
        reform_labels=["CB=£50", "CB=£300"],
    )



from random import random
import numpy as np
import tensorflow as tf
from openfisca_uk import parameters
from openfisca_uk.calibration.losses.loss_category import LossCategory


class PopulationsByAgeSexRegion(LossCategory):
    label = "Population by age, sex and region"
    parameter_folder = parameters.calibration.populations.by_age_sex_region

    def get_loss_subcomponents(
        sim,
        household_weights,
        year,
    ):
        person_sex = sim.calc("gender").values
        person_age = sim.calc("age").values
        person_region = sim.calc("region", map_to="person").values
        population = PopulationsByAgeSexRegion.parameter_folder
        for sex in population.children:
            for region in population.children[sex].children:
                for age_group in (
                    population.children[sex].children[region].children
                ):
                    parameter = (
                        population.children[sex]
                        .children[region]
                        .children[age_group]
                    )
                    parameter_name = parameter.name + "." + str(year)
                    if "BETWEEN" in age_group:
                        _, lower, upper = age_group.split("_")
                        lower, upper = float(lower), float(upper)
                    elif "OVER" in age_group:
                        lower, upper = float(age_group.split("_")[1]), np.inf
                    else:
                        raise ValueError(
                            f"Unexpected test group: {parameter_name}"
                        )
                    people_in_household = sim.map_to(
                        (person_sex == sex)
                        * (person_age >= lower)
                        * (person_age < upper)
                        * (person_region == region),
                        "person",
                        "household",
                    )
                    model_population = tf.reduce_sum(
                        people_in_household * household_weights
                    )
                    actual_population = parameter(f"{year}-01-01")
                    if people_in_household.sum() > 0:
                        # If the FRS has no observations, skip the target.
                        yield parameter_name, model_population, actual_population



from openfisca_uk.model_api import *


class bi_phaseout(Variable):
    label = "Basic income phase-out"
    entity = Person
    definition_period = YEAR
    value_type = float
    unit = "currency-GBP"

    def formula(person, period, parameters):
        income = person("total_income", period)
        bi = parameters(period).contrib.ubi_center.basic_income
        max_bi = person("bi_maximum", period)
        income_over_threshold = max_(income - bi.phase_out.threshold, 0)
        uncapped_deduction = bi.phase_out.rate * income_over_threshold
        return min_(max_bi, uncapped_deduction)



from openfisca_uk_data import FRSEnhanced
from openfisca_uk import Microsimulation
from openfisca_tools.testing import generate_tests


test_parameter = generate_tests(
    Microsimulation(dataset=FRSEnhanced, year=2019)
)



from openfisca_uk.calibration.losses.categories.budgetary_impact import (
    BudgetaryImpact,
)
from openfisca_uk.calibration.losses.categories.families import (
    Families,
)
from openfisca_uk.calibration.losses.categories.households import (
    Households,
)
from openfisca_uk.calibration.losses.categories.populations import (
    Populations,
)



from openfisca_uk.model_api import *


class AFCS(Variable):
    value_type = float
    entity = Person
    label = "Armed Forces Compensation Scheme"
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(person, period, parameters):
        return person("AFCS_reported", period)


class AFCS_reported(Variable):
    value_type = float
    entity = Person
    label = "Armed Forces Compensation Scheme (reported)"
    definition_period = YEAR
    unit = "currency-GBP"



from openfisca_uk.system import CountryTaxBenefitSystem


class variable_holder(type):
    contents = {
        variable.name: type(variable)
        for variable in CountryTaxBenefitSystem().variables.values()
    }

    def __getattr__(cls, name):
        return cls.contents[name]


class variables(metaclass=variable_holder):
    pass



from openfisca_uk.model_api import *


class working_tax_credit_reported(Variable):
    value_type = float
    entity = Person
    label = "Working Tax Credit"
    definition_period = YEAR
    unit = "currency-GBP"


class child_tax_credit_reported(Variable):
    value_type = float
    entity = Person
    label = "Working Tax Credit"
    definition_period = YEAR
    unit = "currency-GBP"


class tax_credits_applicable_income(Variable):
    value_type = float
    entity = BenUnit
    label = "Applicable income for Tax Credits"
    definition_period = YEAR
    unit = "currency-GBP"
    reference = "The Tax Credits (Definition and Calculation of Income) Regulations 2002 s. 3"

    def formula(benunit, period, parameters):
        TC = parameters(period).benefit.tax_credits
        STEP_1_COMPONENTS = [
            "pension_income",
            "savings_interest_income",
            "dividend_income",
            "property_income",
        ]
        income = aggr(benunit, period, STEP_1_COMPONENTS)
        income = amount_over(income, TC.means_test.non_earned_disregard)
        STEP_2_COMPONENTS = [
            "employment_income",
            "self_employment_income",
            "social_security_income",
            "miscellaneous_income",
        ]
        bi = parameters(period).contrib.ubi_center.basic_income
        if bi.include_in_means_tests:
            STEP_2_COMPONENTS.append("basic_income")
        income += aggr(benunit, period, STEP_2_COMPONENTS)
        EXEMPT_BENEFITS = ["income_support", "ESA_income", "JSA_income"]
        on_exempt_benefits = add(benunit, period, EXEMPT_BENEFITS) > 0
        return income * ~on_exempt_benefits


class is_CTC_child_limit_exempt(Variable):
    value_type = bool
    entity = Person
    label = "Exemption from Child Tax Credit child limit"
    documentation = "Exemption from Child Tax Credit limit on number of children based on birth year"
    definition_period = YEAR

    def formula(person, period, parameters):
        limit_year = parameters(
            period
        ).benefit.tax_credits.child_tax_credit.limit.start_year
        # Children must be born before April 2017.
        # We use < 2017 as the closer approximation than <= 2017.
        return person("birth_year", period) < limit_year


class is_child_for_CTC(Variable):
    value_type = bool
    entity = Person
    label = "Child eligible for Child Tax Credit"
    definition_period = YEAR
    reference = "Tax Credits Act 2002 s. 8"

    def formula(person, period, parameters):
        return person("is_child_or_QYP", period)


class is_CTC_eligible(Variable):
    value_type = bool
    entity = BenUnit
    label = "Child Tax Credit eligibility"
    definition_period = YEAR
    reference = "Tax Credits Act 2002 s. 8"

    def formula(benunit, period, parameters):
        already_claiming = (
            aggr(benunit, period, ["child_tax_credit_reported"]) > 0
        )
        return (
            benunit.any(benunit.members("is_child_for_CTC", period))
            & already_claiming
        )


class would_claim_CTC(Variable):
    value_type = bool
    entity = BenUnit
    label = "Would claim Child Tax Credit"
    documentation = (
        "Whether this family would claim Child Tax Credit if eligible"
    )
    definition_period = YEAR

    def formula(benunit, period, parameters):
        reported_ctc = aggr(benunit, period, ["child_tax_credit_reported"]) > 0
        claims_all_entitled_benefits = benunit(
            "claims_all_entitled_benefits", period
        )
        baseline = benunit("baseline_has_working_tax_credit", period)
        takeup_rate = parameters(period).benefit.housing_benefit.takeup
        return select(
            [reported_ctc | claims_all_entitled_benefits, ~baseline, True],
            [
                True,
                random(benunit) < takeup_rate,
                False,
            ],
        )


class CTC_maximum_rate(Variable):
    value_type = float
    entity = BenUnit
    label = "Maximum Child Tax Credit"
    definition_period = YEAR
    reference = "Tax Credits Act 2002 s. 9"
    unit = "currency-GBP"

    def formula(benunit, period, parameters):
        ELEMENTS = [
            "CTC_family_element",
            "CTC_child_element",
            "CTC_disabled_child_element",
            "CTC_severely_disabled_child_element",
        ]
        return add(benunit, period, ELEMENTS)


class CTC_family_element(Variable):
    value_type = float
    entity = BenUnit
    label = "CTC entitlement in the Family Element"
    definition_period = YEAR
    reference = "Tax Credits Act 2002 s. 9"
    unit = "currency-GBP"

    def formula(benunit, period, parameters):
        CTC = parameters(period).benefit.tax_credits.child_tax_credit
        return benunit("is_CTC_eligible", period) * CTC.elements.family_element


class CTC_child_element(Variable):
    value_type = float
    entity = BenUnit
    label = "Child Tax Credit child element"
    definition_period = YEAR
    reference = "Tax Credits Act 2002 s. 9"
    unit = "currency-GBP"

    def formula(benunit, period, parameters):
        person = benunit.members
        CTC = parameters(period).benefit.tax_credits.child_tax_credit
        is_child_for_CTC = person("is_child_for_CTC", period)
        is_CTC_child_limit_exempt = person("is_CTC_child_limit_exempt", period)
        exempt_child = is_child_for_CTC & is_CTC_child_limit_exempt
        exempt_children = benunit.sum(exempt_child)
        child_limit = CTC.limit.child_count
        spaces_left = max_(0, child_limit - exempt_children)
        non_exempt_children = min_(spaces_left, benunit.sum(is_child_for_CTC))
        children = exempt_children + non_exempt_children
        amount = CTC.elements.child_element * children
        return amount


class CTC_disabled_child_element(Variable):
    value_type = float
    entity = BenUnit
    label = "CTC entitlement from disabled child elements"
    definition_period = YEAR
    reference = "Tax Credits Act 2002 s. 9"
    unit = "currency-GBP"

    def formula(benunit, period, parameters):
        person = benunit.members
        is_child_for_CTC = person("is_child_for_CTC", period)
        is_disabled_for_benefits = person("is_disabled_for_benefits", period)
        is_disabled_child = is_child_for_CTC & is_disabled_for_benefits
        disabled_children = benunit.sum(is_disabled_child)
        CTC = parameters(period).benefit.tax_credits.child_tax_credit
        amount = CTC.elements.dis_child_element * disabled_children
        return benunit("is_CTC_eligible", period) * amount


class CTC_severely_disabled_child_element(Variable):
    value_type = float
    entity = BenUnit
    label = "CTC entitlement from severely disabled child elements"
    definition_period = YEAR
    reference = "Tax Credits Act 2002 s. 9"
    unit = "currency-GBP"

    def formula(benunit, period, parameters):
        person = benunit.members
        is_child_for_CTC = person("is_child_for_CTC", period)
        is_severely_disabled_for_benefits = person(
            "is_severely_disabled_for_benefits", period
        )
        is_severely_disabled_child = (
            is_child_for_CTC & is_severely_disabled_for_benefits
        )
        severely_disabled_children = benunit.sum(is_severely_disabled_child)
        CTC = parameters(period).benefit.tax_credits.child_tax_credit
        amount = (
            CTC.elements.severe_dis_child_element * severely_disabled_children
        )
        return benunit("is_CTC_eligible", period) * amount


class is_WTC_eligible(Variable):
    value_type = bool
    entity = BenUnit
    label = "Working Tax Credit eligibility"
    definition_period = YEAR
    reference = "Tax Credits Act 2002 s. 10"

    def formula(benunit, period, parameters):
        WTC = parameters(period).benefit.tax_credits.working_tax_credit
        person = benunit.members
        person_hours = person("weekly_hours", period)
        total_hours = benunit.sum(person_hours)
        max_person_hours = benunit.max(person_hours)
        has_disabled_adults = benunit("num_disabled_adults", period) > 0
        family_type = benunit("family_type", period)
        families = family_type.possible_values
        old = person("age", period.this_year) >= WTC.min_hours.old_age
        has_old = benunit.any(old)
        lone_parent = family_type == families.LONE_PARENT
        couple_with_children = family_type == families.COUPLE_WITH_CHILDREN
        eldest_25_plus = benunit("eldest_adult_age", period) >= 25
        youngest_under_60 = benunit("youngest_adult_age", period) < 60
        # Calculate WTC eligibility group.
        lower_req = has_disabled_adults | has_old | lone_parent
        medium_req = couple_with_children & ~lower_req
        higher_req = eldest_25_plus & youngest_under_60
        # Calculate eligibility for each WTC group.
        meets_lower = total_hours >= WTC.min_hours.lower
        meets_medium_total_hours = (
            total_hours >= WTC.min_hours.couple_with_children
        )
        meets_medium_person_hours = max_person_hours >= WTC.min_hours.lower
        meets_medium = meets_medium_total_hours & meets_medium_person_hours
        meets_higher = total_hours >= WTC.min_hours.default
        already_claiming = (
            aggr(benunit, period, ["working_tax_credit_reported"]) > 0
        )
        return (
            (lower_req & meets_lower)
            | (medium_req & meets_medium)
            | (higher_req & meets_higher)
        ) & already_claiming


class would_claim_WTC(Variable):
    value_type = bool
    entity = BenUnit
    label = "Would claim Working Tax Credit"
    documentation = (
        "Whether this family would claim Working Tax Credit if eligible"
    )
    definition_period = YEAR

    def formula(benunit, period, parameters):
        reported_wtc = (
            aggr(benunit, period, ["working_tax_credit_reported"]) > 0
        )
        claims_all_entitled_benefits = benunit(
            "claims_all_entitled_benefits", period
        )
        baseline = benunit("baseline_has_child_tax_credit", period)
        takeup_rate = parameters(period).benefit.housing_benefit.takeup
        return select(
            [reported_wtc | claims_all_entitled_benefits, ~baseline, True],
            [
                True,
                random(benunit) < takeup_rate,
                False,
            ],
        )


class WTC_maximum_rate(Variable):
    value_type = float
    entity = BenUnit
    label = "Working Tax Credit maximum rate"
    definition_period = YEAR
    reference = "Tax Credits Act 2002 s. 11"
    unit = "currency-GBP"

    def formula(benunit, period, parameters):
        ELEMENTS = [
            "WTC_basic_element",
            "WTC_couple_element",
            "WTC_lone_parent_element",
            "WTC_disabled_element",
            "WTC_severely_disabled_element",
            "WTC_worker_element",
            "WTC_childcare_element",
        ]
        return add(benunit, period, ELEMENTS)


class WTC_basic_element(Variable):
    value_type = float
    entity = BenUnit
    label = "Working Tax Credit basic element"
    definition_period = YEAR
    reference = "Tax Credits Act 2002 s. 11"
    unit = "currency-GBP"

    def formula(benunit, period, parameters):
        WTC = parameters(period).benefit.tax_credits.working_tax_credit
        return benunit("is_WTC_eligible", period) * WTC.elements.basic


class WTC_couple_element(Variable):
    value_type = float
    entity = BenUnit
    label = "Working Tax Credit couple element"
    definition_period = YEAR
    reference = "Tax Credits Act 2002 s. 11"
    unit = "currency-GBP"

    def formula(benunit, period, parameters):
        WTC = parameters(period).benefit.tax_credits.working_tax_credit
        relation_type = benunit("relation_type", period)
        relations = relation_type.possible_values
        amount = (relation_type == relations.COUPLE) * WTC.elements.couple
        return benunit("is_WTC_eligible", period) * amount


class WTC_lone_parent_element(Variable):
    value_type = float
    entity = BenUnit
    label = "Working Tax Credit lone parent element"
    definition_period = YEAR
    reference = "Tax Credits Act 2002 s. 11"
    unit = "currency-GBP"

    def formula(benunit, period, parameters):
        WTC = parameters(period).benefit.tax_credits.working_tax_credit
        family_type = benunit("family_type", period)
        families = family_type.possible_values
        lone_parent = family_type == families.LONE_PARENT
        return (
            benunit("is_WTC_eligible", period)
            * lone_parent
            * WTC.elements.lone_parent
        )


class WTC_disabled_element(Variable):
    value_type = float
    entity = BenUnit
    label = "Working Tax Credit disabled element"
    definition_period = YEAR
    reference = "Tax Credits Act 2002 s. 11"
    unit = "currency-GBP"

    def formula(benunit, period, parameters):
        WTC = parameters(period).benefit.tax_credits.working_tax_credit
        person = benunit.members
        person_meets_hours = (
            person("weekly_hours", period) >= WTC.min_hours.lower
        )
        person_qualifies = (
            person_meets_hours
            & person("is_disabled_for_benefits", period)
            & person("is_adult", period)
        )
        qualifies = benunit.any(person_qualifies)
        return (
            benunit("is_WTC_eligible", period)
            * qualifies
            * WTC.elements.disabled
        )


class WTC_severely_disabled_element(Variable):
    value_type = float
    entity = BenUnit
    label = "Working Tax Credit severely disabled element"
    definition_period = YEAR
    reference = "Tax Credits Act 2002 s. 11"
    unit = "currency-GBP"

    def formula(benunit, period, parameters):
        WTC = parameters(period).benefit.tax_credits.working_tax_credit
        amount = (
            benunit("num_severely_disabled_adults", period)
            * WTC.elements.severely_disabled
        )
        return benunit("is_WTC_eligible", period) * amount


class WTC_worker_element(Variable):
    value_type = float
    entity = BenUnit
    label = "Working Tax Credit worker element"
    definition_period = YEAR
    reference = "Tax Credits Act 2002 s. 11"
    unit = "currency-GBP"

    def formula(benunit, period, parameters):
        WTC = parameters(period).benefit.tax_credits.working_tax_credit
        hours = aggr(benunit, period, ["weekly_hours"])
        meets_hours_requirement = hours >= WTC.min_hours.default
        return (
            benunit("is_WTC_eligible", period)
            * meets_hours_requirement
            * WTC.elements.worker
        )


class WTC_childcare_element(Variable):
    value_type = float
    entity = BenUnit
    label = "Working Tax Credit childcare element"
    definition_period = YEAR
    reference = "Tax Credits Act 2002 s. 11"
    unit = "currency-GBP"

    def formula(benunit, period, parameters):
        WTC = parameters(period).benefit.tax_credits.working_tax_credit
        num_children = benunit("num_children", period)
        childcare_1 = (num_children == 1) * WTC.elements.childcare_1
        childcare_2 = (num_children > 1) * WTC.elements.childcare_2
        max_childcare_amount = (childcare_1 + childcare_2) * WEEKS_IN_YEAR
        expenses = aggr(benunit, period, ["childcare_expenses"])
        eligible_expenses = min_(max_childcare_amount, expenses)
        childcare_element = WTC.elements.childcare_coverage * eligible_expenses
        return benunit("is_WTC_eligible", period) * childcare_element


class tax_credits_reduction(Variable):
    value_type = float
    entity = BenUnit
    label = "Reduction in Tax Credits from means-tested income"
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(benunit, period, parameters):
        means_test = parameters(period).benefit.tax_credits.means_test
        CTC_amount = benunit("CTC_maximum_rate", period)
        WTC_amount = benunit("WTC_maximum_rate", period)
        CTC_only = (CTC_amount > 0) & (WTC_amount == 0)
        threshold = where(
            CTC_only,
            means_test.income_threshold_CTC_only,
            means_test.income_threshold,
        )
        income = benunit("tax_credits_applicable_income", period)
        overage = max_(0, income - threshold)
        return overage * means_test.income_reduction_rate


class working_tax_credit_pre_minimum(Variable):
    value_type = float
    entity = BenUnit
    label = "Working Tax Credit pre-minimum"
    documentation = (
        "Working Tax Credit amount before the minimum tax credit is applied"
    )
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(benunit, period, parameters):
        return max_(
            0,
            benunit("WTC_maximum_rate", period)
            - benunit("tax_credits_reduction", period),
        )


class child_tax_credit_pre_minimum(Variable):
    value_type = float
    entity = BenUnit
    label = "Child Tax Credit pre-minimum"
    documentation = (
        "Child Tax Credit amount before the minimum tax credit is applied"
    )
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(benunit, period, parameters):
        reduction_left = max_(
            0,
            benunit("tax_credits_reduction", period)
            - benunit("WTC_maximum_rate", period),
        )
        return max_(
            0,
            benunit("CTC_maximum_rate", period) - reduction_left,
        )


class tax_credits(Variable):
    value_type = float
    entity = BenUnit
    label = "Tax Credits"
    documentation = "Value of the Tax Credits (benefits) for this family"
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(person, period, parameters):
        amount = add(
            person,
            period,
            ["working_tax_credit_pre_minimum", "child_tax_credit_pre_minimum"],
        )
        min_benefit = parameters(period).benefit.tax_credits.min_benefit
        return where(amount < min_benefit, 0, amount)


class child_tax_credit(Variable):
    value_type = float
    entity = BenUnit
    label = "Child Tax Credit"
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(benunit, period, parameters):
        return where(
            benunit("tax_credits", period) > 0,
            benunit("child_tax_credit_pre_minimum", period),
            0,
        ) * (
            benunit("is_CTC_eligible", period)
            & benunit("would_claim_CTC", period)
        )


class working_tax_credit(Variable):
    value_type = float
    entity = BenUnit
    label = "Working Tax Credit"
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(benunit, period, parameters):
        return where(
            benunit("tax_credits", period) > 0,
            benunit("working_tax_credit_pre_minimum", period),
            0,
        ) * (
            benunit("is_WTC_eligible", period)
            & benunit("would_claim_WTC", period)
        )


class baseline_has_working_tax_credit(Variable):
    label = "Receives Working Tax Credit (baseline)"
    entity = BenUnit
    definition_period = YEAR
    value_type = bool
    default_value = True


class baseline_has_child_tax_credit(Variable):
    label = "Receives Child Tax Credit (baseline)"
    entity = BenUnit
    definition_period = YEAR
    value_type = bool
    default_value = True



from openfisca_uk.graphs import mtr_chart, budget_chart
from openfisca_uk import BASELINE_PARAMETERS
from openfisca_uk.api import reforms, graphs, BASELINE_PARAMETERS


def test_wide_form_individual_data_with_one_reform():
    abolish_PA = reforms.structural.abolish("personal_allowance")
    graphs.data.get_wide_reform_individual_data(
        [abolish_PA], ["Abolish PA"], ["net_income", "tax"]
    )


def test_wide_form_individual_data_with_two_reforms():
    abolish_PA = reforms.structural.abolish("personal_allowance")
    halve_PA = reforms.parametric.set_parameter(
        BASELINE_PARAMETERS.tax.income_tax.allowances.personal_allowance.amount,
        6250,
    )
    graphs.data.get_wide_reform_individual_data(
        [halve_PA, abolish_PA],
        ["Halve PA", "Abolish PA"],
        ["net_income", "tax"],
    )


def test_single_reform_MTR_and_budget_plot():
    reform = reforms.structural.abolish(
        "personal_allowance"
    ), reforms.parametric.set_parameter(
        BASELINE_PARAMETERS.benefit.universal_credit.means_test.reduction_rate,
        0.4,
    )
    budget_chart(reform)
    mtr_chart(reform)


def test_multiple_reform_MTR_and_budget_plot():
    reform_1 = reforms.structural.abolish(
        "personal_allowance"
    ), reforms.parametric.set_parameter(
        BASELINE_PARAMETERS.benefit.universal_credit.means_test.reduction_rate,
        0.4,
    )
    reform_2 = reforms.structural.abolish(
        "personal_allowance"
    ), reforms.parametric.set_parameter(
        BASELINE_PARAMETERS.benefit.universal_credit.means_test.reduction_rate,
        0.1,
    )
    budget_chart([reform_1, reform_2], names=["CB=£50", "CB=£300"])
    mtr_chart([reform_1, reform_2], names=["CB=£50", "CB=£300"])



from openfisca_uk.model_api import *


class income_support_reported(Variable):
    value_type = float
    entity = Person
    label = "Income Support (reported amount)"
    definition_period = YEAR
    unit = "currency-GBP"


class would_claim_IS(Variable):
    value_type = bool
    entity = BenUnit
    label = "Would claim Income Support"
    documentation = (
        "Whether this family would claim Income Support if eligible"
    )
    definition_period = YEAR

    def formula(benunit, period, parameters):
        reported_is = aggr(benunit, period, ["income_support_reported"]) > 0
        claims_all_entitled_benefits = benunit(
            "claims_all_entitled_benefits", period
        )
        baseline = benunit("baseline_has_income_support", period)
        takeup_rate = parameters(period).benefit.housing_benefit.takeup
        return select(
            [reported_is | claims_all_entitled_benefits, ~baseline, True],
            [
                True,
                random(benunit) < takeup_rate,
                False,
            ],
        )


class baseline_has_income_support(Variable):
    label = "Receives Income Support (baseline)"
    entity = BenUnit
    definition_period = YEAR
    value_type = bool
    default_value = True


class income_support_applicable_income(Variable):
    value_type = float
    entity = BenUnit
    label = "Relevant income for Income Support means test"
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(benunit, period, parameters):
        IS = parameters(period).benefit.income_support
        INCOME_COMPONENTS = [
            "employment_income",
            "self_employment_income",
            "property_income",
            "pension_income",
        ]
        bi = parameters(period).contrib.ubi_center.basic_income
        if bi.include_in_means_tests:
            INCOME_COMPONENTS.append("basic_income")
        income = aggr(benunit, period, INCOME_COMPONENTS)
        tax = aggr(
            benunit,
            period,
            ["income_tax", "national_insurance"],
        )
        income += aggr(benunit, period, ["social_security_income"])
        income -= tax
        income -= aggr(benunit, period, ["pension_contributions"]) * 0.5
        family_type = benunit("family_type", period)
        families = family_type.possible_values
        # Calculate income disregards for each family type.
        mt = IS.means_test
        single = family_type == families.SINGLE
        income_disregard_single = single * mt.income_disregard_single
        single = family_type == families.SINGLE
        income_disregard_couple = (
            benunit("is_couple", period) * mt.income_disregard_couple
        )
        lone_parent = family_type == families.LONE_PARENT
        income_disregard_lone_parent = (
            lone_parent * mt.income_disregard_lone_parent
        )
        income_disregard = (
            income_disregard_single
            + income_disregard_couple
            + income_disregard_lone_parent
        ) * WEEKS_IN_YEAR
        return max_(0, income - income_disregard)


class income_support_eligible(Variable):
    value_type = bool
    entity = BenUnit
    label = "Whether eligible for Income Support"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        youngest_child_5_or_under = benunit("youngest_child_age", period) <= 5
        lone_parent = benunit("is_lone_parent", period)
        lone_parent_with_young_child = lone_parent & youngest_child_5_or_under
        has_carers = aggr(benunit, period, ["is_carer_for_benefits"]) > 0
        none_SP_age = ~benunit.any(benunit.members("is_SP_age", period))
        has_ESA_income = benunit("ESA_income", period) > 0
        already_claiming = (
            aggr(benunit, period, ["income_support_reported"]) > 0
        )
        return (
            (has_carers | lone_parent_with_young_child)
            & none_SP_age
            & ~has_ESA_income
            & already_claiming
        )


class income_support_applicable_amount(Variable):
    value_type = float
    entity = BenUnit
    label = "Applicable amount of Income Support"
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(benunit, period, parameters):
        IS = parameters(period).benefit.income_support
        amounts = IS.amounts
        younger_age = benunit("youngest_adult_age", period)
        older_age = benunit("eldest_adult_age", period)
        younger_under_18 = younger_age < 18
        younger_under_25 = younger_age < 25
        older_under_18 = older_age < 18
        has_children = benunit.any(benunit.members("is_child", period))
        single = benunit("is_single", period)
        single_under_25 = single & ~has_children & younger_under_25
        single_over_25 = single & ~has_children & ~younger_under_25
        lone_young = single & has_children & younger_under_18
        lone_old = single & has_children & ~younger_under_18
        couple_young = ~single & older_under_18
        couple_mixed = ~single & ~older_under_18 & younger_under_18
        couple_old = ~single & ~younger_under_18
        personal_allowance_weekly = select(
            [
                single_under_25,
                single_over_25,
                lone_young,
                lone_old,
                couple_young,
                couple_mixed,
                couple_old,
            ],
            [
                amounts.amount_16_24,
                amounts.amount_over_25,
                amounts.amount_lone_16_17,
                amounts.amount_lone_over_18,
                amounts.amount_couples_16_17,
                amounts.amount_couples_age_gap,
                amounts.amount_couples_over_18,
            ],
        )
        personal_allowance = personal_allowance_weekly * WEEKS_IN_YEAR
        premiums = benunit("benefits_premiums", period)
        return (
            (personal_allowance + premiums)
            * benunit("income_support_eligible", period)
            * benunit("would_claim_IS", period)
        )


class income_support(Variable):
    value_type = float
    entity = BenUnit
    label = "Income Support"
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(benunit, period, parameters):
        amount = benunit("income_support_applicable_amount", period)
        income = benunit("income_support_applicable_income", period)
        return max_(0, amount - income)



from openfisca_uk.model_api import *


class CB_HITC(Variable):
    value_type = float
    entity = Person
    label = "Child Benefit High-Income Tax Charge"
    definition_period = YEAR
    reference = "https://www.legislation.gov.uk/ukpga/2003/1/part/10/chapter/8"
    unit = "currency-GBP"

    def formula(person, period, parameters):
        CB_received = person.benunit("child_benefit", period)
        CB_HITC = parameters(period).tax.income_tax.charges.CB_HITC
        percentage = (
            amount_over(
                person("adjusted_net_income", period),
                CB_HITC.phase_out_start,
            )
            # HITC is specified as a percent of Child Benefit recollected for
            # every £1,000 over the phase-out threshold.
            / 1_000
            * CB_HITC.phase_out_rate
        )
        is_higher_earner = person("is_higher_earner", period)
        return (min_(percentage, 1) * CB_received) * is_higher_earner



import tensorflow as tf
from openfisca_uk import parameters
from openfisca_uk.calibration.losses.loss_category import LossCategory


class HouseholdsByRegionByCouncilTaxBand(LossCategory):
    label = "Households by region and council tax band"
    parameter_folder = (
        parameters.calibration.households.by_region_by_council_tax_band
    )

    def get_loss_subcomponents(
        sim,
        household_weights,
        year,
    ):
        total_households = parameters.calibration.households.in_total(
            f"{year}-01-01"
        )
        total_tenure_metrics = sum(
            [
                parameter(f"{year}-01-01")
                for parameter in HouseholdsByRegionByCouncilTaxBand.get_metrics()
            ]
        )
        adjustment = total_households / total_tenure_metrics
        ct_bands = HouseholdsByRegionByCouncilTaxBand.parameter_folder
        ct_band = sim.calc("council_tax_band")
        hh_region = sim.calc("region")
        for region in ct_bands.children:
            for band in ct_bands.children[region].children:
                parameter = ct_bands.children[region].children[band]
                parameter_name = parameter.name + "." + str(year)
                household_in_condition = (
                    (hh_region == region) * (ct_band == band)
                ).values
                model_population = tf.reduce_sum(
                    household_weights * household_in_condition
                )
                actual_population = parameter(f"{year}-01-01") * adjustment
                if household_in_condition.sum() > 0:
                    # If the FRS has no observations, skip the target.
                    yield parameter_name, model_population, actual_population



from openfisca_uk.model_api import *
from openfisca_uk.tools.variables import variables


def apply_takeup_rates() -> Type[Reform]:
    """Creates a Reform object applying benefit takeup rates to means-tested benefits.

    Returns:
        Type[Reform]: The OpenFisca reform.
    """

    class would_claim_CTC(variables.would_claim_CTC):
        def formula(benunit, period, parameters):
            return (
                random(benunit)
                < parameters(
                    period
                ).benefit.tax_credits.child_tax_credit.takeup
            )

    class would_claim_WTC(variables.would_claim_WTC):
        def formula(benunit, period, parameters):
            return (
                random(benunit)
                < parameters(
                    period
                ).benefit.tax_credits.working_tax_credit.takeup
            )

    class would_claim_HB(variables.would_claim_HB):
        def formula(benunit, period, parameters):
            return (
                random(benunit)
                < parameters(period).benefit.housing_benefit.takeup
            )

    class would_claim_IS(variables.would_claim_IS):
        def formula(benunit, period, parameters):
            return (
                random(benunit)
                < parameters(period).benefit.income_support.takeup
            )

    class would_claim_PC(variables.would_claim_PC):
        def formula(benunit, period, parameters):
            return (
                random(benunit)
                < parameters(period).benefit.pension_credit.takeup
            )

    class claims_legacy_benefits(variables.claims_legacy_benefits):
        def formula(benunit, period, parameters):
            # Assign legacy/UC claimant status, consistently for each household
            household = benunit.members.household
            benunit_random = benunit.value_from_first_person(
                household.project(random(household))
            )
            UC_rollout = parameters(
                period
            ).benefit.universal_credit.rollout_rate
            return benunit_random > UC_rollout

    class takeup_rates(Reform):
        def apply(self):
            self.update_variable(would_claim_CTC)
            self.update_variable(would_claim_WTC)
            self.update_variable(would_claim_HB)
            self.update_variable(would_claim_IS)
            self.update_variable(would_claim_PC)
            self.update_variable(claims_legacy_benefits)

    return takeup_rates



from openfisca_uk.calibration.losses.categories.households.in_total import (
    HouseholdsInTotal,
)
from openfisca_uk.calibration.losses.categories.households.by_region_by_council_tax_band import (
    HouseholdsByRegionByCouncilTaxBand,
)
from openfisca_uk.calibration.losses.categories.households.by_region_by_tenure_type import (
    HouseholdsByRegionByTenureType,
)
from openfisca_uk.calibration.losses.loss_category import (
    combine_loss_categories,
)

Households = combine_loss_categories(
    HouseholdsInTotal,
    HouseholdsByRegionByCouncilTaxBand,
    HouseholdsByRegionByTenureType,
    label="Households",
)



import numpy as np
from openfisca_uk.calibration.losses.loss_category import LossCategory
import tensorflow as tf
from openfisca_uk import parameters
from openfisca_core.parameters import Parameter


class PopulationsByAgeBand(LossCategory):
    label = "Populations by age band"
    parameter_folder = parameters.calibration.populations.by_age_sex_region

    def get_loss_subcomponents(
        sim,
        household_weights,
        year,
    ):
        total_population = parameters.calibration.populations.in_total(
            f"{year}-01-01"
        )
        total_age_bands = sum(
            [
                parameter(f"{year}-01-01")
                for parameter in PopulationsByAgeBand.parameter_folder.get_descendants()
                if isinstance(parameter, Parameter)
            ]
        )
        adjustment = total_population / total_age_bands
        population = PopulationsByAgeBand.parameter_folder
        person_age = sim.calc("age", year)
        age_groups = {}
        for sex in population.children:
            for region in population.children[sex].children:
                for age_group in (
                    population.children[sex].children[region].children
                ):
                    parameter = (
                        population.children[sex]
                        .children[region]
                        .children[age_group]
                    )
                    if age_group not in age_groups:
                        age_groups[age_group] = parameter(f"{year}-01-01")
                    else:
                        age_groups[age_group] += parameter(f"{year}-01-01")
        for age_group in age_groups:
            if "BETWEEN" in age_group:
                _, lower, upper = age_group.split("_")
                lower, upper = int(lower), int(upper)
            elif "OVER" in age_group:
                lower, upper = int(age_group.split("_")[1]), np.inf
                if lower == 80:
                    lower = 79  # The FRS top-codes at 79, so we'll be very slightly mismatched
                    # but it's better than dropping over-80 targeting altogether
            else:
                raise ValueError(f"Unexpected test group: {age_group}")
            people_in_household = sim.map_to(
                (person_age >= lower) * (person_age < upper),
                "person",
                "household",
            )
            model_population = tf.reduce_sum(
                people_in_household * household_weights
            )
            if people_in_household.sum() > 0:
                # If the FRS has no observations, skip the target.
                yield getattr(
                    PopulationsByAgeBand.parameter_folder.MALE.LONDON,
                    age_group,
                ).name + "." + str(year), model_population, age_groups[
                    age_group
                ] * adjustment

    def get_metrics():
        return (
            PopulationsByAgeBand.parameter_folder.MALE.LONDON.get_descendants()
        )

    def get_metric_names():
        return [
            parameter.name + "." + str(year)
            for year in range(2019, 2023)
            for parameter in PopulationsByAgeBand.get_metrics()
        ]



from typing import List
import plotly.express as px
from openfisca_uk.api import *
import pandas as pd
from microdf import MicroSeries
from functools import wraps
from openfisca_uk.graphs.general import net_cost


def _get_group_reduction(
    incomes: MicroSeries,
    baseline_values: MicroSeries,
    reform_values: MicroSeries,
    bucket: str,
    aggregation: str,
    compare_groups: bool,
    relative: bool,
):
    def reduction_function(values):
        return values.groupby(
            incomes.__getattribute__(f"{bucket}_rank")()
        ).__getattribute__(aggregation)()

    if compare_groups:
        change = reduction_function(reform_values) - reduction_function(
            baseline_values
        )
        if relative:
            change /= reduction_function(incomes)
    else:
        if not relative:
            change = reduction_function(reform_values - baseline_values)
        else:
            change = reduction_function(
                (reform_values - baseline_values) / incomes
            )
    return change


def distributional_chart(
    reforms: List[Reform],
    bucket_variable: str,
    change_variable: str,
    level: str = "person",
    reform_names: List[str] = None,
    bucket: str = "decile",
    aggregation: str = "mean",
    compare_groups: bool = "true",
    relative: bool = False,
    currency: str = "£",
    include_all_ticks: bool = True,
    baseline: Microsimulation = None,
    microsimulation_kwargs: dict = {},
    chart_type: str = "bar",
    **kwargs,
):
    """Generates a Plotly bar chart for changes to a variable based on quantiles.

    Args:
        reforms (List[Reform]): Reform or list of reforms to calculate effects for.
        bucket_variable (str): The variable to group entities by.
        change_variable (str): The variable to show changes in.
        level (str, optional): The entity level to map to. Defaults to "person".
        reform_names (List[str], optional): Names of reforms, if multiple passed.
        bucket (str, optional): The type of quantile grouping (quintile, decile, percentile). Defaults to "decile".
        value_label (str, optional): The name of the value variable. Defaults to "net income".
        aggregation (str, optional): The type of aggregation (mean, sum, median). Defaults to "mean".
        compare_groups (bool, optional): Whether to compare group aggregates (true) or aggregate individual comparisons (false).
        relative (bool, optional): Whether changes should be a percentage of their baseline value.
        currency (str, optional): The currency format to use if using absolute changes.
        baseline (Microsimulation, optional): Baseline simulation, can improve speed.
        microsimulation_kwargs (dict, optional). Keyword arguments to pass when constructing reform simulations. Defaults to {}.
        **kwargs (dict): Any additional arguments to pass to fig.update_layout (e.g. title=...).

    Returns:
        Figure: The Plotly figure.
    """
    assert bucket in (
        "quintile",
        "decile",
        "percentile",
    ), "Unrecognised bucketing type"
    assert aggregation in (
        "mean",
        "sum",
        "median",
    ), "Unrecognised aggregation type"
    assert chart_type in (
        "bar",
        "line",
    ), "Unrecognised chart type"
    x_label = f"{BASELINE_VARIABLES[bucket_variable].label} {bucket}"
    y_label = f"Change to {BASELINE_VARIABLES[change_variable].label.lower()}"

    baseline = baseline or Microsimulation()
    bucket_values = baseline.calc(bucket_variable, map_to=level)
    baseline_values = baseline.calc(change_variable, map_to=level)

    def reduction_func(reform_values):
        return _get_group_reduction(
            bucket_values,
            baseline_values,
            reform_values,
            bucket,
            aggregation,
            compare_groups,
            relative,
        )

    if isinstance(reforms, list):
        dfs = []
        for reform, name in zip(reforms, reform_names):
            reform_sim = Microsimulation(reform, **microsimulation_kwargs)
            result = reduction_func(
                reform_sim.calc(change_variable, map_to=level)
            )
            df = pd.DataFrame(
                {x_label: result.index, y_label: result.values, "Reform": name}
            )
            dfs += [df]
        data = pd.concat(dfs)
    else:
        reform_sim = Microsimulation(reforms, **microsimulation_kwargs)
        result = reduction_func(reform_sim.calc(change_variable, map_to=level))
        data = pd.DataFrame(
            {x_label: result.index, y_label: result.values, "Reform": "Reform"}
        )

    if chart_type == "bar":
        fig = px.bar(data, x=x_label, y=y_label, animation_frame="Reform")
    elif chart_type == "line":
        fig = px.line(data, x=x_label, y=y_label, animation_frame="Reform")
    if relative:
        fig.update_layout(yaxis_tickformat="%")
    else:
        fig.update_layout(yaxis_tickprefix=currency)
    if include_all_ticks:
        numbers = dict(quartile=4, quintile=5, decile=10, percentile=100)[
            bucket
        ]
        if numbers <= 10:
            ticks = list(range(1, numbers + 1))
        else:
            ticks = None
        fig.update_layout(xaxis_tickvals=ticks)
    fig.update_layout(**kwargs)
    return fig


# Bucket-specific shortcut functions

quartile_chart = wraps(distributional_chart)(
    lambda *args, **kwargs: distributional_chart(
        *args, bucket="quartile", **kwargs
    )
)
quintile_chart = wraps(distributional_chart)(
    lambda *args, **kwargs: distributional_chart(
        *args, bucket="quintile", **kwargs
    )
)
decile_chart = wraps(distributional_chart)(
    lambda *args, **kwargs: distributional_chart(
        *args, bucket="decile", **kwargs
    )
)
percentile_chart = wraps(distributional_chart)(
    lambda *args, **kwargs: distributional_chart(
        *args, bucket="percentile", **kwargs
    )
)


def waterfall_data(
    reforms: List[Reform],
    reform_labels: List[str],
    subreform_labels: List[str],
    baseline: Microsimulation = None,
    aggregate: str = "net_income",
    invert_y: bool = False,
    **kwargs,
):
    """Generates data for a waterfall funding breakdown chart.

    Args:
        reforms (List[Reform]): A list of reforms to show funding breakdowns for. Each reform must have the same top-level subreforms.
        reform_labels (List[str]): The names of the reforms.
        subreform_labels (List[str]): The names of the components of the reforms.
        baseline (Microsimulation, optional): A baseline simulation - can improve speed if already initialised. Defaults to None.
        aggregate (str): The variable to use as the aggregate cost variable. Defaults to "net_income".
        invert_y (bool): Whether the invert the y axis (y == revenue or y == spending). Defaults to False.

    Returns:
        pd.DataFrame: The waterfall plot dataframe.
    """
    baseline = baseline or Microsimulation(**kwargs)
    dfs = []
    for reform, name in zip(reforms, reform_labels):
        net_costs = []
        for i in range(len(reform) + 1):
            net_costs += [
                net_cost(
                    baseline,
                    Microsimulation(reform[:i], **kwargs),
                    invert=invert_y,
                    variable=aggregate,
                )
            ]
        net_costs = np.array(net_costs)
        labels = subreform_labels + ["Remaining"]
        resulting_costs = pd.Series(net_costs[1:] - net_costs[:-1])
        resulting_costs = pd.Series(list(resulting_costs) + [net_costs[-1]])
        if net_costs[-1] < 0:
            final_base = 0
        else:
            final_base = -net_costs[-1]
        breakdown = pd.DataFrame(
            dict(
                # Each component has white below it, so double the records.
                component=labels * 2,
                amount=[0]
                + list(
                    np.where(
                        resulting_costs[1:-1] < 0,
                        -resulting_costs.cumsum()[:-2],
                        -resulting_costs.cumsum()[1:-1],
                    )
                )
                + [final_base]
                + list(resulting_costs.abs()),
                Type=["-"] * len(labels)
                + np.where(
                    resulting_costs < 0, "Revenue", "Spending"
                ).tolist(),
                Reform=name,
            )
        )
        dfs += [breakdown]
    return pd.concat(dfs)


SPENDING_COLOR = "#BDBDBD"
REVENUE_COLOR = "#1976D2"


def waterfall_chart(
    reforms: List[Reform],
    subreform_labels: List[str],
    reform_labels: List[str] = None,
    baseline: Microsimulation = None,
    aggregate: str = "net_income",
    invert_y: bool = False,
):
    """Generates a waterfull funding breakdown plot.

    Args:
        reforms (Union[Reform, List[Reform]]): A list of reforms - each must have the same top-level subreforms.
        subreform_labels (List[str]): The names of the top-level reform components.
        reform_labels (List[str], optional): The names of the reforms, if multiple are provided.
        baseline (Microsimulation, optional): A baseline simulation - can improve speed if pre-initialised. Defaults to None.
        aggregate (str): The variable to use as the aggregate cost variable. Defaults to "net_income".
        invert_y (bool): Whether the invert the y axis (y == revenue or y == spending). Defaults to False.

    Returns:
        Figure: The Plotly figure.
    """
    if not isinstance(reforms, list):
        reforms = [reforms]
        reform_labels = ["Reform"]

    return px.bar(
        waterfall_data(
            reforms,
            reform_labels,
            subreform_labels,
            baseline=baseline,
            aggregate=aggregate,
            invert_y=invert_y,
        ),
        x="component",
        y="amount",
        color="Type",
        barmode="stack",
        animation_frame="Reform",
        color_discrete_sequence=["white", REVENUE_COLOR, SPENDING_COLOR],
    ).update_layout(
        xaxis_title="Component",
        yaxis_title="Surplus",
        yaxis_tickprefix="£",
        legend_title="",
        # White background to match the unaffected area.
        paper_bgcolor="rgba(0,0,0,0)",
        plot_bgcolor="rgba(0,0,0,0)",
    )



import numpy as np
from openfisca_uk.calibration.losses.loss_category import LossCategory
import tensorflow as tf
from openfisca_uk import parameters


class BudgetaryImpactOfIncomeTaxByIncomeBand(LossCategory):
    label = "Income tax revenue by income"
    parameter_folder = (
        parameters.calibration.budgetary_impact.of_income_tax_by_income_band
    )

    def get_loss_subcomponents(
        sim,
        household_weights,
        year,
    ):
        brackets = (
            BudgetaryImpactOfIncomeTaxByIncomeBand.parameter_folder.brackets
        )  # Last-3 brackets are impossible due to FRS non-capture
        num_thresholds = len(brackets)
        instant_str = f"{year}-01-01"
        for i in range(num_thresholds):
            lower_threshold = brackets[i].threshold(instant_str)
            upper_threshold = (
                brackets[i + 1].threshold(instant_str)
                if i < num_thresholds - 1
                else np.inf
            )
            income = sim.calc("total_income", period=year).values
            person_in_range = (income >= lower_threshold) & (
                income < upper_threshold
            )
            income_tax_in_range = (
                sim.calc("income_tax", period=year).values * person_in_range
            )
            household_income_tax = sim.map_to(
                income_tax_in_range, "person", "household"
            )
            aggregate = tf.reduce_sum(household_weights * household_income_tax)
            target = brackets[i].amount(instant_str)
            if person_in_range.sum() > 0:
                # If the FRS does not have any observations, skip the target.
                yield brackets[i].name + "." + str(year), aggregate, target

    def get_metrics():
        return (
            BudgetaryImpactOfIncomeTaxByIncomeBand.parameter_folder.brackets[
                :-3
            ]
        )

    def get_metric_names():
        return [
            bracket.name + "." + str(year)
            for bracket in BudgetaryImpactOfIncomeTaxByIncomeBand.parameter_folder.brackets[
                :-3
            ]
            for year in range(2019, 2023)
        ]



from openfisca_uk.calibration.losses.categories.budgetary_impact.by_program import (
    BudgetaryImpactByProgram,
)
from openfisca_uk.calibration.losses.categories.budgetary_impact.by_income_source import (
    BudgetaryImpactByIncomeSource,
)
from openfisca_uk.calibration.losses.categories.budgetary_impact.by_program_by_country import (
    BudgetaryImpactByProgramByCountry,
)
from openfisca_uk.calibration.losses.categories.budgetary_impact.of_income_tax_by_income_band import (
    BudgetaryImpactOfIncomeTaxByIncomeBand,
)
from openfisca_uk.calibration.losses.loss_category import (
    combine_loss_categories,
)

BudgetaryImpact = combine_loss_categories(
    BudgetaryImpactByProgram,
    BudgetaryImpactByIncomeSource,
    BudgetaryImpactByProgramByCountry,
    BudgetaryImpactOfIncomeTaxByIncomeBand,
    label="Budgetary impact",
    weight=3,
)



from openfisca_uk import Microsimulation
import pytest
from openfisca_uk_data import FRSEnhanced

YEARS = range(2019, 2022)

baseline = Microsimulation(dataset=FRSEnhanced, year=2019)


@pytest.mark.parametrize("year", YEARS)
def test_not_nan(year):
    for variable in baseline.simulation.tax_benefit_system.variables:
        assert ~baseline.calc(variable, period=year).isna().any()



# -*- coding: utf-8 -*-

import os
from openfisca_uk import entities
from openfisca_uk.system import CountryTaxBenefitSystem
from openfisca_uk.tools.simulation import (
    IndividualSim,
    Microsimulation,
    prepare_parameters,
)
from openfisca_uk.reforms.presets.modelling import (
    reported_tax,
    reported_benefits,
    reported,
)
from pathlib import Path
import os
from openfisca_core.taxbenefitsystems import TaxBenefitSystem

COUNTRY_DIR = os.path.dirname(os.path.abspath(__file__))

system = prepare_parameters(CountryTaxBenefitSystem())
parameters = BASELINE_PARAMETERS = system.parameters


class AttributeDict(dict):
    __slots__ = ()
    __getattr__ = dict.__getitem__
    __setattr__ = dict.__setitem__


BASELINE_VARIABLES = AttributeDict(
    {name: value.__class__ for name, value in system.variables.items()}
)
variables = BASELINE_VARIABLES
REPO = Path(__file__).parent

# Our country tax and benefit class inherits from the general TaxBenefitSystem class.
# The name CountryTaxBenefitSystem must not be changed, as all tools of the OpenFisca ecosystem expect a CountryTaxBenefitSystem class to be exposed in the __init__ module of a country package.



from openfisca_uk.model_api import *


class housing_benefit_reported(Variable):
    value_type = float
    entity = Person
    label = "Housing Benefit (reported amount)"
    definition_period = YEAR
    unit = "currency-GBP"


class housing_benefit_eligible(Variable):
    value_type = bool
    entity = BenUnit
    label = "Whether eligible for Housing Benefit"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        social = benunit.any(benunit.members("in_social_housing", period))
        already_claiming = (
            aggr(benunit, period, ["housing_benefit_reported"]) > 0
        )
        return already_claiming & (social | benunit("LHA_eligible", period))


class would_claim_HB(Variable):
    value_type = bool
    entity = BenUnit
    label = "Would claim Housing Benefit"
    documentation = (
        "Whether this family would claim Housing Benefit if eligible"
    )
    definition_period = YEAR

    def formula(benunit, period, parameters):
        claims_all_entitled_benefits = benunit(
            "claims_all_entitled_benefits", period
        )
        reported_hb = aggr(benunit, period, ["housing_benefit_reported"]) > 0
        baseline = benunit("baseline_has_housing_benefit", period)
        takeup_rate = parameters(period).benefit.housing_benefit.takeup
        return select(
            [reported_hb | claims_all_entitled_benefits, ~baseline, True],
            [
                True,
                random(benunit) < takeup_rate,
                False,
            ],
        )


class baseline_has_housing_benefit(Variable):
    label = "Receives Housing Benefit (baseline)"
    entity = BenUnit
    definition_period = YEAR
    value_type = bool
    default_value = True


class housing_benefit_applicable_amount(Variable):
    value_type = float
    entity = BenUnit
    label = "Applicable amount for Housing Benefit"
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(benunit, period, parameters):
        HB = parameters(period).benefit.housing_benefit
        PA = HB.allowances
        any_over_SP_age = benunit.any(benunit.members("is_SP_age", period))
        eldest_age = benunit("eldest_adult_age", period)
        u_18 = eldest_age < 18
        u_25 = eldest_age < 25
        o_25 = (eldest_age >= 25) & ~any_over_SP_age
        o_18 = (eldest_age >= 18) * ~any_over_SP_age
        single = benunit("is_single_person", period)
        couple = benunit("is_couple", period)
        lone_parent = benunit("is_lone_parent", period)
        single_personal_allowance = (
            u_25 * PA.single.under_25
            + o_25 * PA.single.over_25
            + any_over_SP_age * PA.single.SP_age
        )
        couple_personal_allowance = (
            u_18 * PA.couple.both_under_18
            + o_18 * PA.couple.over_18
            + any_over_SP_age * PA.couple.SP_age
        )
        lone_parent_personal_allowance = (
            u_18 * PA.lone_parent.under_18
            + o_18 * PA.lone_parent.over_18
            + any_over_SP_age * PA.lone_parent.SP_age
        )
        personal_allowance = (
            single * single_personal_allowance
            + couple * couple_personal_allowance
            + lone_parent * lone_parent_personal_allowance
        ) * WEEKS_IN_YEAR
        premiums = benunit("benefits_premiums", period)
        housing_benefit_eligible = benunit("housing_benefit_eligible", period)
        return (personal_allowance + premiums) * housing_benefit_eligible


class housing_benefit_applicable_income(Variable):
    value_type = float
    entity = BenUnit
    label = "Relevant income for Housing Benefit means test"
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(benunit, period, parameters):
        WTC = parameters(period).benefit.tax_credits.working_tax_credit
        means_test = parameters(period).benefit.housing_benefit.means_test
        BENUNIT_MEANS_TESTED_BENEFITS = [
            "child_benefit",
            "income_support",
            "JSA_income",
            "ESA_income",
        ]
        INCOME_COMPONENTS = [
            "employment_income",
            "self_employment_income",
            "property_income",
            "pension_income",
        ]
        bi = parameters(period).contrib.ubi_center.basic_income
        TAX_COMPONENTS = ["income_tax", "national_insurance"]
        benefits = add(benunit, period, BENUNIT_MEANS_TESTED_BENEFITS)
        income = aggr(benunit, period, INCOME_COMPONENTS)
        tax = aggr(benunit, period, TAX_COMPONENTS)
        income += aggr(benunit, period, ["personal_benefits"])
        if not bi.include_in_means_tests:
            # Basic income is already in personal benefits, deduct if needed
            income -= add(benunit, period, ["basic_income"])
        income += add(benunit, period, ["tax_credits"])
        income -= tax
        income -= aggr(benunit, period, ["pension_contributions"]) * 0.5
        income += benefits
        num_children = benunit.nb_persons(BenUnit.CHILD)
        childcare_amount_1 = (num_children == 1) * WTC.elements.childcare_1
        childcare_amount_2 = (num_children > 1) * WTC.elements.childcare_2
        max_weekly_childcare_amount = childcare_amount_1 + childcare_amount_2
        max_childcare_amount = max_weekly_childcare_amount * WEEKS_IN_YEAR
        childcare_element = min_(
            max_childcare_amount,
            aggr(benunit, period, ["childcare_expenses"]),
        )
        hours = aggr(benunit, period, ["weekly_hours"])
        # Calculate single, couple, lone parent, and worker disregards.
        single = benunit("is_single_person", period)
        single_disregard = single * means_test.income_disregard_single
        couple = benunit("is_couple", period)
        couple_disregard = couple * means_test.income_disregard_couple
        lone_parent = benunit("is_lone_parent", period)
        lone_parent_disregard = (
            lone_parent * means_test.income_disregard_lone_parent
        )
        hour_requirement = where(
            lone_parent, WTC.min_hours.lower, means_test.worker_hours
        )
        worker = hours > hour_requirement
        worker_disregard = worker * means_test.worker_income_disregard
        weekly_disregard = (
            single_disregard
            + couple_disregard
            + lone_parent_disregard
            + worker_disregard
        )
        disregard = weekly_disregard * WEEKS_IN_YEAR
        return max_(0, income - disregard - childcare_element)


class HB_individual_non_dep_deduction(Variable):
    value_type = float
    entity = Person
    label = "Non-dependent deduction (individual)"
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(person, period, parameters):
        not_rent_liable = person.benunit("benunit_rent", period) == 0
        over_21 = person("age", period) >= 21
        deduction_scale = parameters(
            period
        ).benefit.housing_benefit.deductions.non_dep_deduction
        weekly_income = person("total_income", period)
        deduction = deduction_scale.calc(weekly_income)
        return deduction * over_21 * not_rent_liable * MONTHS_IN_YEAR


class HB_non_dep_deductions(Variable):
    value_type = float
    entity = BenUnit
    label = "Non-dependent deductions"
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(benunit, period, parameters):
        non_dep_deductions_in_hh = benunit.max(
            benunit.members.household.sum(
                benunit.members("HB_individual_non_dep_deduction", period)
            )
        )
        non_dep_deductions_in_bu = aggr(
            benunit, period, ["HB_individual_non_dep_deduction"]
        )
        return non_dep_deductions_in_hh - non_dep_deductions_in_bu


class housing_benefit(Variable):
    value_type = float
    entity = BenUnit
    label = "Housing Benefit"
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(benunit, period, parameters):
        rent = benunit("benunit_rent", period)
        LHA = benunit("LHA_eligible", period.this_year)
        applicable_amount = benunit(
            "housing_benefit_applicable_amount", period
        )
        income = benunit("housing_benefit_applicable_income", period)
        withdrawal_rate = parameters(
            period
        ).benefit.housing_benefit.means_test.withdrawal_rate
        final_amount = max_(
            0, rent - max_(0, income - applicable_amount) * withdrawal_rate
        )
        amount = where(
            LHA, min_(final_amount, benunit("LHA_cap", period)), final_amount
        )
        CAPPED_BENUNIT_BENEFITS = [
            "child_benefit",
            "child_tax_credit",
            "JSA_income",
            "income_support",
            "ESA_income",
        ]
        capped_benunit_benefits = add(benunit, period, CAPPED_BENUNIT_BENEFITS)
        CAPPED_PERSONAL_BENEFITS = [
            "JSA_contrib",
            "incapacity_benefit",
            "ESA_contrib",
            "SDA",
        ]
        capped_personal_benefits = aggr(
            benunit, period, CAPPED_PERSONAL_BENEFITS
        )
        other_capped_benefits = (
            capped_benunit_benefits + capped_personal_benefits
        )
        amount = max_(0, amount - benunit("HB_non_dep_deductions", period))
        final_amount = min_(
            amount
            * (
                benunit("housing_benefit_eligible", period)
                & benunit("would_claim_HB", period)
            ),
            benunit("benefit_cap", period) - other_capped_benefits,
        )
        return max_(0, final_amount)



from openfisca_core.parameters import ParameterNode, Parameter
from openfisca_core.periods import instant


def add_tax_benefit_uprating(parameters: ParameterNode) -> ParameterNode:
    """Adds the uprating index used to uprate most tax-benefit instruments (September CPI of the previous year)."""
    cpi = parameters.uprating.CPI
    data = {}
    for year in range(2000, 2029):
        data[f"{year + 1}-04-01"] = cpi(instant(f"{year}-09-01"))
    if not hasattr(parameters.uprating, "september_cpi"):
        parameters.uprating.add_child(
            "september_cpi", Parameter("september_cpi", data=data)
        )
    return parameters



from openfisca_uk.calibration.losses.loss_category import LossCategory
import tensorflow as tf
from openfisca_uk import parameters


class HouseholdsInTotal(LossCategory):
    label = "Households in total"
    parameter_folder = parameters.calibration.households.in_total

    def get_loss_subcomponents(
        sim,
        household_weights,
        year,
    ):
        model_population = tf.reduce_sum(household_weights)
        actual_population = HouseholdsInTotal.parameter_folder(f"{year}-01-01")
        yield HouseholdsInTotal.parameter_folder.name + "." + str(
            year
        ), model_population, actual_population

    def get_metrics():
        return [HouseholdsInTotal.parameter_folder]

    def get_metric_names():
        return [
            HouseholdsInTotal.parameter_folder.name + "." + str(year)
            for year in range(2019, 2023)
        ]



from openfisca_uk.model_api import *


class PIP(Variable):
    value_type = float
    entity = Person
    label = "Personal Independence Payment"
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(person, period, parameters):
        return add(person, period, ["PIP_M", "PIP_DL"])


class PIP_DL(Variable):
    value_type = float
    entity = Person
    label = "Personal Independence Payment (Daily Living)"
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(person, period, parameters):
        return person("PIP_DL_reported", period)


class PIP_M(Variable):
    value_type = float
    entity = Person
    label = "Personal Independence Payment (Mobility)"
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(person, period, parameters):
        return person("PIP_M_reported", period)


class PIP_DL_reported(Variable):
    value_type = float
    entity = Person
    label = "Personal Independence Payment (Daily Living) (reported)"
    definition_period = YEAR
    unit = "currency-GBP"


class PIP_M_reported(Variable):
    value_type = float
    entity = Person
    label = "Personal Independence Payment (Mobility) (reported)"
    definition_period = YEAR
    unit = "currency-GBP"



from openfisca_uk import Microsimulation
from openfisca_uk_data import FRSEnhanced
import pytest
from itertools import product

UPRATED_VARIABLES = (
    "council_tax",
    "employment_income",
    "pension_income",
    "self_employment_income",
    "property_income",
    "savings_interest_income",
    "dividend_income",
)

sim = Microsimulation(dataset=FRSEnhanced, year=2019)


@pytest.mark.parametrize(
    "variable,year", product(UPRATED_VARIABLES, range(2020, 2022))
)
def test_uprating(variable: str, year: int):
    assert any(
        sim.calc(variable, period=year).values
        != sim.calc(variable, period=2019).values
    )



import argparse
import logging
import os
import sys
from openfisca_uk.tools.testing import run_tests
from openfisca_core.scripts import build_tax_benefit_system

OPENFISCA_UK = "openfisca_uk"


def add_tax_benefit_system_arguments(parser):
    parser.add_argument(
        "-c",
        "--country-package",
        default=OPENFISCA_UK,
        action="store",
        help='country package to use. If not provided, an automatic detection will be attempted by scanning the python packages installed in your environment which name contains the word "openfisca".',
    )
    parser.add_argument(
        "-e",
        "--extensions",
        default=[],
        action="store",
        help="extensions to load",
        nargs="*",
    )
    parser.add_argument(
        "-r",
        "--reforms",
        action="store",
        help="reforms to apply to the country package",
        nargs="*",
    )

    return parser


def build_parser():
    parser = argparse.ArgumentParser()

    subparsers = parser.add_subparsers(
        help="Available commands", dest="command"
    )
    subparsers.required = (
        True  # Can be added as an argument of add_subparsers in Python 3
    )

    def build_test_parser(parser):
        parser = add_tax_benefit_system_arguments(parser)
        parser.add_argument(
            "path",
            help="paths (files or directories) of tests to execute",
            nargs="+",
        )
        parser.add_argument(
            "-n",
            "--name_filter",
            default=None,
            help="partial name of tests to execute. Only tests with the given name_filter in their name, file name, or keywords will be run.",
        )
        parser.add_argument(
            "-p",
            "--pdb",
            action="store_true",
            default=False,
            help="drop into debugger on failures or errors",
        )
        parser.add_argument(
            "--performance-graph",
            "--performance",
            action="store_true",
            default=False,
            help="output a performance graph in a 'performance_graph.html' file",
        )
        parser.add_argument(
            "--performance-tables",
            action="store_true",
            default=False,
            help="output performance CSV tables",
        )
        parser.add_argument(
            "-v",
            "--verbose",
            action="store_true",
            default=False,
            help="increase output verbosity",
        )
        parser.add_argument(
            "-o",
            "--only-variables",
            nargs="*",
            default=None,
            help="variables to test. If specified, only test the given variables.",
        )
        parser.add_argument(
            "-i",
            "--ignore-variables",
            nargs="*",
            default=None,
            help="variables to ignore. If specified, do not test the given variables.",
        )

        return parser

    parser_test = subparsers.add_parser(
        "test", help="Run OpenFisca YAML tests"
    )
    parser_test = build_test_parser(parser_test)

    return parser


def main():
    args = build_parser().parse_args()
    logging.basicConfig(
        level=logging.DEBUG if args.verbose else logging.WARNING,
        stream=sys.stdout,
    )

    tax_benefit_system = build_tax_benefit_system(
        OPENFISCA_UK, args.extensions, args.reforms
    )

    options = {
        "pdb": args.pdb,
        "performance_graph": args.performance_graph,
        "performance_tables": args.performance_tables,
        "verbose": args.verbose,
        "name_filter": args.name_filter,
        "only_variables": args.only_variables,
        "ignore_variables": args.ignore_variables,
    }

    paths = [os.path.abspath(path) for path in args.path]
    sys.exit(run_tests(tax_benefit_system, paths, options))


if __name__ == "__main__":
    main()
